$Script:ErrorActionPreference = "Stop"
$Script:DebugPreference = "Continue"

<#
.SYNOPSIS
Backs up a VM.

.DESCRIPTION
Creates a backup job in Veeam and submits the specified VM into this job. A request is then made for this job to begin immediately. If a job name is not specified, a job name will be created in the format {vm_name}-{vm_remote_id}-mm/dd/yyyy hh:mm:ss {AM|PM}. If there is another backup job with the same name, an error is thrown. Unique job names are recommended.

You can specify a number of V-Commander custom attribute parameters to retrieve post-backup information such as the last successful backup timestamp, the number of restore points and the job name that was created or used. You must create these custom attributes in V-Commander before running this command. This command will perform an actual backup, which could take anywhere from minutes to potentially hours. V-Commander custom attributes will be updated when there is a successful backup. 


.EXAMPLE
Backup-VM -computerName 'prod.mycompany.com' -vCommanderCredential 'C:\\Scripts\vcmdr_encrypted_credential_key.xml' -remoteComputerName 'veeam.mycompany.com' -veeamCredential 'C:\\Scripts\veeam_encrypted_credential_key.xml' -hostName 'esx1.embotics.com' -vmName 'VM001' -vmRemoteId 'vm-11538'  -jobName 'VM001-vm-11538-backup'

This command backs up a VM without retrieving any post-backup metadata from Veeam.

.EXAMPLE
Backup-VM -computerName 'prod.mycompany.com' -vCommanderCredential 'C:\\Scripts\vcmdr_encrypted_credential_key.xml' -remoteComputerName 'veeam.mycompany.com' -veeamCredential 'C:\\Scripts\veeam_encrypted_credential_key.xml' -hostName 'esx1.embotics.com' -vmName 'VM001' -vmRemoteId 'vm-11538'  -jobName 'VM001-vm-11538-backup' -attribNameLastBackup 'LastBackupTime' -attribNameJobName 'JobName' -attribNameNumRestorePoints 'NumRestorePoints'

This command backs up a VM, retrieves metadata from Veeam and stores the metadata in custom attributes for this VM: "LastBackupTime", "JobName" and "NumRestorePoints".
#>

function Backup-VM {
    [CmdletBinding()]
	param(
        [string] $computerName = $(Throw "Provide the computer name of the V-Commander server."),
        [String] $vCommanderCredential = $(Throw "Provide the credential to access V-Commander."),
        [string] $remoteComputerName = $(Throw "Provide the computer name of the Veeam server."), 
        [String] $veeamCredential = $(Throw "Provide the credential to access Veeam server."),
        [string] $hostName = $(Throw "Provide the host where the VM is located at."), 
        [string] $vmName = $(Throw "Provide the VM name."), 
        [string] $vmRemoteId = $(Throw "Provide the remote ID of this VM."),
		[string] $jobName, #the job name to backup the VM. If this is null, we will create for the user
        [string] $attribNameLastBackup, # the custom attribute to store the timestamp of the last successful backup
        [string] $attribNameJobName, # the custom attribute to store the job name
        [string] $attribNameNumRestorePoints # the custom attribute to store the # of restore points
        
    )

	if (-not $PSBoundParameters['Debug']) {
		$PSBoundParameters['Debug'] = $false
	}

	if ($PSBoundParameters['Debug']) {$DebugPreference = "Continue"}
	if ($Global:DebugPreference -eq "Continue") {
		$DebugPreference="Continue"
	}

	Set-PSDebug -Strict
	Set-StrictMode version Latest
    
    if([string]::IsNullOrEmpty($computerName)) {
        Write-Error "The computer name of the V-Commander server was not specified or is empty."
        Throw The computer name of the V-Commander server was not specified or is empty.
    }
   
    if([string]::IsNullOrEmpty($remoteComputerName)) {
        Write-Error "The computer name of the Veeam server was not specified or is empty."
        Throw The computer name of the Veeam server was not specified or is empty.
    }
    
	$index = $computerName.IndexOf(":")
	if ($index -eq -1) {
		$computeNameNoPort = $computerName
	} else{
		$computeNameNoPort = $computerName.Substring(0,$index)
	}
	
    Write-Debug "Checking connectivity to $computerName"
    $vcmdrConnectionResult = Test-Connection -computer $computeNameNoPort -quiet
    if (!$vcmdrConnectionResult) {
        Write-Error "The V-Commander server $computerName could not be reached."
        Throw The V-Commander server $computerName could not be reached. Connectivity to the specified server could not be established.
    }  
    
	$index = $remoteComputerName.IndexOf(":")
	if ($index -eq -1) {
		$computeNameNoPort = $remoteComputerName
	} else{
		$computeNameNoPort = $remoteComputerName.Substring(0,$index)
	}
    Write-Debug "Checking connectivity to $remoteComputerName"
    $veeamServerConnectionResult = Test-Connection -computer $computeNameNoPort -quiet
    if (!$veeamServerConnectionResult) {
        Write-Error "The Veeam server $remoteComputerName could not be reached."
        Throw The Veeam server $remoteComputerName could not be reached. Connectivity to the specified server could not be established.
    }
   
    Write-Debug "Checking if V-Commander credential file exists."
    if (!(Test-Path $vCommanderCredential)) {
        Write-Error "V-Commander credential file does not exists at $vCommanderCredential"
        Throw "V-Commander credential file does not exist at $vCommanderCredential. Please check if this file exists at this location.";
    }
    
    Write-Debug "Checking if Veeam credential file exists."
    if (!(Test-Path $veeamCredential)) {
        Write-Error "Veeam credential file does not exist at $veeamCredential"
        Throw "Veeam credential file does not exists at $veeamCredential. Please check if this file exists at this location.";
    }
    
    import-module VCommander
    
    #Decrypt the V-Commander credential to see if it's valid; if it's not valid, there is no point in contacting Veeam
    #We could avoid checking this if no custom attributes are specified, then there is no need to do this
    Write-Debug "Checking for V-Commander credential validity."
    $vcmdrCred = New-DecryptCredential -keyFilePath $vCommanderCredential

    #Decrypt the credentials to access Veeam server
    Write-Debug "Checking for Veam credential validity."
    $cred = New-DecryptCredential -keyFilePath $veeamCredential

	Write-Debug "host: $hostName | VM Name: $vmName | VM Remote Id: $vmRemoteId | jobName: $jobName"    
    try {
        #We could start a remote session or invoke a command with the current user's credentials (without specifying the -Credential parameter) but let's keep it simple and making this mandatory    
        $invokeResult = invoke-Command -ErrorAction Stop -ComputerName $remoteComputerName -Credential $cred -Argumentlist $hostName, $vmName, $vmRemoteId, $jobName, $PSBoundParameters['Debug'] -scriptblock {
        		param ($hostName, $vmName, $vmRemoteId, $jobName, $debug)
                if ($debug) {$DebugPreference = "Continue"}  
            
        		Add-PSSnapin -name VeeamPSSnapIn
                
                #If a job name is not specified, we will create one for the user
        		if ([string]::IsNullOrEmpty($jobName)) {
                  $backupdate = get-date	
        		  $backupdate = $backupdate.ToString()

        		  write-Debug "Current timestamp: $backupdate"
        		  $jobName = "$vmName-$vmRemoteId-$backupdate"
                  Write-Host "Job name not specified, defaulting to: $jobName"
        		}		
                
        		write-Debug "Job name: $jobName" 
        		Write-Debug "Querying for host"
        		$serverobject = Get-VBRServer -WarningAction:SilentlyContinue -Name $hostName
                if ($serverobject -eq $null) {                
                    Write-Error "Host '$hostName' does not exist or has not been added to Veeam."
                    throw "Host '$hostName' does not exist or has not been added to Veeam. Please verify this host exists or added to Veeam.";
                }

        		Write-Debug "Querying for VM"
        		$vm_entity = Find-VBRViEntity -WarningAction:SilentlyContinue -Name $vmName | where {$_.Reference -eq $vmRemoteId}
                if ($vm_entity -eq $null) {
                    Write-Error "VM '$vmName' with remote id '$vmRemoteId' does not exist."
                    throw "VM '$vmName' with remote id '$vmRemoteId' does not exist. Please verify that this VM exists in one of the vCenters managed by Veeam.";
                }

                Write-Debug "Looking up job: $jobName"
                $JobObject = Get-VBRJob -WarningAction:SilentlyContinue -Name $jobName
                if ($JobObject) { #if it exists
                    if ( ($jobObject | measure).Count -gt 1) {
                        #throw error as we can't determine which job to run
                        Write-Error "Duplicate job name; unable to decide which job to run."
                        throw "Duplicate job name; unable to decide which job to run.";
                    }                    
                    Write-Debug "Existing job found: $jobName"
                } else {
                    Write-Debug "Creating job with name: $jobName"
                    Add-VBRViBackupJob -WarningAction:SilentlyContinue -name $jobName -entity ($vm_entity)
        		    $JobObject = Get-VBRJob -WarningAction:SilentlyContinue -Name $jobName
                }
                
        		Write-Host "Starting job '$jobName'"
        		$jobResult = start-VBRJob -WarningAction:SilentlyContinue $JobObject
                if ($jobResult -eq $null) {
                    Write-Error "Unable to start job '$jobName'. See Veeam log for details."
                    throw "Unable to start job '$jobName'. See Veeam log for details.";
                }

                Write-Debug "Getting # of restore points for: $jobName"
        		$restorePoints = Get-VBRRestorePoint -WarningAction:SilentlyContinue -Backup (Get-VBRBackup -WarningAction:SilentlyContinue -name $jobName)
                if ($restorePoints -eq $null) {
                    Write-Error "Unable to determine the # of restore points for job '$jobName'. See Veeam log for details."
                    throw "Unable to determine the # of restore points for job '$jobName'. See Veeam log for details.";
                }
                
        		$numRestorePoints = $restorePoints | measure        		
        		$returnObject = new-object psobject -property @{ resultText=$jobResult.Result.toString();
        								 endTime=$jobResult.EndTime.toString(); 
        								 jobName=$jobResult.jobName.toString();
        								 numRestorePoints = $numRestorePoints.Count.toString()
        								}
                                        
                Write-Debug "Backup result: $($returnObject.resultText)"                        
                Write-Debug "Backup completed timestamp: $($returnObject.endTime)"
                Write-Debug "Backup job name: $($returnObject.jobName)"
                Write-Debug "Total # of restore points: $($returnObject.numRestorePoints)"
                                        
        		$returnObject
            }    
        } catch {
            Write-Error "$($_.Exception.Message)" 
            throw $_.Exception
        } 
    
    #We are only interested in the last result object; if there are other objects that got sent to the output, we ignore them.
    if ( ($invokeResult | measure ).Count -gt 1 ) {
        $invokeResult = $invokeResult[-1]
    }
	
	#Warning is considered a success. ie. low disk space
    if ( ($invokeResult.resultText -ne 'Success') -and ($invokeResult.resultText -ne 'Warning')) {
        Write-Error "Backup requested failed. Status: $($invokeResult.resultText)" 
        throw "Backup request failed. Status: $($invokeResult.resultText)"
    }
     
    #Create two arrays to store custom attribute names and values                         
    [String[]] $attributeNames = @()
    [String[]] $attributeValues = @()
    
    #If last backup timestamp attribute is specified
    if (!([string]::IsNullOrEmpty($attribNameLastBackup))) {
          $attributeNames += $attribNameLastBackup
          $attributeValues += $invokeResult.endTime
    }
    
    #If job name attribute is specified
    if (!([string]::IsNullOrEmpty($attribNameJobName))) {
          $attributeNames += $attribNameJobName
          $attributeValues += $invokeResult.jobName
    }
    
    #If number of restore points attribute is specified
    if (!([string]::IsNullOrEmpty($attribNameNumRestorePoints))) {
          $attributeNames += $attribNameNumRestorePoints
          $attributeValues += $invokeResult.numRestorePoints
    }
    
    if ($attributeNames.length -gt 0) {
        #Set a set of custom attributes on the VM
        Write-Debug "$($attributeNames.length) custom attributes, updating them."
        Set-VcmdrVmAttribute -computerName $computerName -vcmdrCredentials $vcmdrCred -vmName $vmName -vmRemoteId $vmRemoteId -attributeNames $attributeNames -attributeValues $attributeValues
    } else {
        Write-Debug "No custom attributes to update - skipping."
    }
    
    Write-Host "VM '$vmName' (remoteid: $vmRemoteId) backed up on $($invokeResult.endTime)"
}

#Set the Vcmdr custom attribute
function Set-VcmdrVmAttribute {
	param(
	
		[ValidateNotNull()]
        [String]$computerName = $(Throw "Provide the computer name of the target host."),
       
	    [ValidateNotNull()]
		[System.Management.Automation.PSCredential] $vcmdrCredentials = $(Throw "Provide a System.Management.Automation.PSCredential object containing the credentials for authentication against the REST service."),
        
		[ValidateNotNull()]
		[string]$vmName = $(Throw "Provide the VM name."),
		
		[ValidateNotNull()]
		[string] $vmRemoteId = $(Throw "Provide the remote ID of this VM."),
        [string[]]$attributeNames = $(Throw "Provide the attribute name to set."),
        [string[]]$attributeValues = $(Throw "Provide the attribute value to set.")
    )

	#Import the required modules
	Write-Debug "Importing VCommanderRestClient module"
	Import-Module -Name VCommanderRestClient

	$Global:BASE_SERVICE_URI = $Global:BASE_SERVICE_URI_PATTERN.Replace("{service_host}",$computerName)
	$Global:CREDENTIAL = $vcmdrCredentials

	Write-Debug "Ignoring untrusted certificate error"
	VCommander\Set-IgnoreSslErrors

	#Start communicating with vCommander REST service
	Write-Debug "Connecting as $($Global:CREDENTIAL.UserName) to $($Global:BASE_SERVICE_URI)"
	
	$connected = Connect-Client
	
	Write-Debug "Connected : $connected"
	
	Write-Debug "Looking for VM with remote Id '$($vmRemoteId)'"
	$vmCollection = Get-VMByRemoteId -vmRemoteId $vmRemoteId
	if ($vmCollection.VirtualMachineCollection -eq $null) {
		throw "Unable to locate VM with remote Id '$($vmRemoteId)'"
	}
	
	Write-Debug "About to apply $($attributeNames.Length) attributes."
	#Create a custom attribute DTO
    $attributeDTO = New-DTOTemplateObject -DTOTagName "CustomAttribute"
    $attributeDTO.CustomAttribute.allowedValues = @()
    $attributeDTO.CustomAttribute.description = $null
    $attributeDTO.CustomAttribute.targetManagedObjectTypes = @()
	for($counter=0; $counter -lt $attributeNames.Length; $counter++) {
		
		$attributeDTO.CustomAttribute.name = $attributeNames[$counter]
        $attributeDTO.CustomAttribute.value = $null
        if (-not [string]::IsNullOrEmpty($attributeValues[$counter])) {
            $attributeDTO.CustomAttribute.value = $attributeValues[$counter]
        }

        Write-Debug "Applying attribute '$($attributeNames[$counter])' with value '$($attributeValues[$counter])'"
		try {
			VCommanderRestClient\Set-Attribute -vmId $vmCollection.VirtualMachineCollection.VirtualMachines[0].id -customAttributeDTo $attributeDTO | Out-null
		} catch  {
			Write-Debug "$vmName : failed to apply custom attribute(s) - skipping."
			throw $_.ErrorDetails.Message
		}
	}
}

<#
.SYNOPSIS
Adds a VM to an existing Veeam backup job.

.DESCRIPTION
Adds a VM to an existing backup job. The backup job must exist at the time of this request or an error is thrown. Once the VM is added to the job, the actual backup occurrence is dependent on the schedule of the job.

.EXAMPLE
Add-VMToJob -remoteComputerName 'veeam.mycompany.com' -veeamCredential 'C:\\Scripts\veeam_encrypted_key.xml' -vmName 'VM001' -vmRemoteId 'vm-11538' -hostName 'esx1.embotics.com' -jobName 'Monthly Backup'

This command adds a VM to an existing Veeam backup job named 'Weekly Backup'.
#>
function Add-VMToJob {
    [CmdletBinding()]
	param(
        [string] $remoteComputerName = $(Throw "Provide the computer name of the Veeam server."), 
        [String] $veeamCredential = $(Throw "Provide the credential to access Veeam server."),
        [string] $hostName = $(Throw "Provide the host where the VM is located at."), 
        [string] $vmName = $(Throw "Provide the VM name."), 
        [string] $vmRemoteId = $(Throw "Provide the remote ID of this VM."),
		[string] $jobName = $(Throw "Enter a job name.")
    )
    if ($PSBoundParameters['Debug']) {$DebugPreference = "Continue"}
    Set-PSDebug -Strict
	Set-StrictMode version Latest
    
    import-module VCommander

	$index = $remoteComputerName.IndexOf(":")
	if ($index -eq -1) {
		$computeNameNoPort = $remoteComputerName
	} else{
		$computeNameNoPort = $remoteComputerName.Substring(0,$index)
	}

    Write-Debug "Checking connectivity to $remoteComputerName"
    $veeamServerConnectionResult = Test-Connection -computer $computeNameNoPort -quiet
    if (!$veeamServerConnectionResult) {
        Write-Error "The Veeam server $remoteComputerName could not be reached."
        Throw The Veeam server $remoteComputerName could not be reached. Connectivity to the specified server could not be established.
    }
      
    Write-Debug "Checking if Veeam credential file exists."
    if (!(Test-Path $veeamCredential)) {
        Write-Error "Veeam credential file does not exist at $veeamCredential"
        Throw "Veeam credential file does not exists at $veeamCredential. Please check if this file exists at this location.";
    }
    
    #Decrypt the credentials to access Veeam server
    $cred = New-DecryptCredential -keyFilePath $veeamCredential
    
    Write-Debug "host: $hostName | VM Name: $vmName | VM Remote Id: $vmRemoteId | jobName: $jobName"      
    try {
        $invokeResult = invoke-Command -ErrorAction Stop -ComputerName $remoteComputerName -Credential $cred -Argumentlist $hostName, $vmName, $vmRemoteId, $jobName, $PSBoundParameters['Debug'] -scriptblock {
    		param ($hostName, $vmName, $vmRemoteId, $jobName, $debug)
            if ($debug) {$DebugPreference = "Continue"}  

    		Add-PSSnapin -name VeeamPSSnapIn
            
             #If a job name is not specified, we will throw error
    		if ([string]::IsNullOrEmpty($jobName)) {
                Write-Error "Job name was not specified." 
                throw "Job name was not specified."
    		}
       	
    		Write-Debug "Querying for host"
    		$serverObject = Get-VBRServer -WarningAction:SilentlyContinue -Name $hostName
            if ($serverobject -eq $null) {                
                Write-Error "Host '$hostName' does not exist or has not been added to Veeam."
                throw "Host '$hostName' does not exist or has not been added to Veeam. Please verify this host exists or added to Veeam."
            }
           
            Write-Debug "Looking up job: $jobName"
            $job = Get-VBRJob -WarningAction:SilentlyContinue -Name $jobName
            if ($job -eq $null) {
                Write-Error "Job '$jobName' does not exist. Please add job to Veeam."
                throw "Job '$jobName' does not exist. Please add job to Veeam."
            }
              
            #Check if VM already in job
            Write-Debug "Checking if VM is already in job"  
            $objectsInJob = Get-VBRJobObject -WarningAction:SilentlyContinue -Job $job -Name $vmName
            $correctObjectInJob = $objectsInJob
            if( ($objectsInJob | measure).Count -gt 0) {  #more than one VM with the same name
                $correctObjectInJob = $objectsInJob | where {$_.Object.ObjectId -eq $vmRemoteId} #find the correct one
            }
            
            if ($correctObjectInJob -ne $null) {
                Write-Error "VM '$vmName' ($vmRemoteId) already exists in job '$jobName'"
                throw "VM '$vmName' ($vmRemoteId) already exists in job '$jobName'"
            }
            
            Write-Debug "Querying for VM"	
            $jobObject = Find-VBRObject -WarningAction:SilentlyContinue -Server $serverObject | where {$_.Id -eq $vmRemoteId}
            if ($jobObject -eq $null) {
                 Write-Error "VM '$vmName' with remote id '$vmRemoteId' does not exist."
                 throw "VM '$vmName' with remote id '$vmRemoteId' does not exist. Please verify that this VM exists in one of the vCenters managed by Veeam.";
            }
           
            Write-Host "Staring to add VM '$vmName' ($vmRemoteId) to job '$jobName'"
            $result = Add-VBRJobObject -WarningAction:SilentlyContinue -Job $job -server $serverobject -entities $jobObject
            if( ($result | measure).Count -gt 0) { #will show all existing VMs in this job as well
                $retrievedVmInfo = $result | where {$_.Object.ObjectId -eq $vmRemoteId}  #check if it contains one of our VMs we just added
                if( ($retrievedVmInfo | measure).Count -gt 0) {
                    Write-Host "VM '$vmName' ($vmRemoteId) was added to job '$jobName'" 
                    return;
                }
            }        
            Write-Error "VM '$vmName' ($vmRemoteId) not add to job '$jobName'" 
        } 
    } catch {
        Write-Error "$($_.Exception.Message)" 
        throw $_.Exception
    }
}

<#
.SYNOPSIS
Removes a VM from an existing backup job.

.DESCRIPTION
Adds this VM to an exclusion list so that it is no longer backed up. The actual removal process depends on the retention policy configured for this job. The backup job must exist or an error is thrown. If the specified VM is already in the exclusion list, the command is ignored.

.EXAMPLE
Remove-VMFromJob  -remoteComputerName 'veeam.mycompany.com' -veeamCredential 'C:\\Scripts\veeam_encrypted_key.xml' -vmName 'VM001' -vmRemoteId 'vm-11538'  -jobName 'Monthly Backup'

This command removes a VM from an existing Veeam backup job named 'Weekly Backup'.
#>
function Remove-VMFromJob {
    [CmdletBinding()]
	param(
        [string] $remoteComputerName = $(Throw "Provide the computer name of the Veeam server."), 
        [String] $veeamCredential = $(Throw "Provide the credential to access Veeam server."),
        [string] $vmName = $(Throw "Provide the VM name."), 
        [string] $vmRemoteId = $(Throw "Provide the remote ID of this VM."),
		[string] $jobName = $(Throw "Enter a job name.")
    ) 
    if ($PSBoundParameters['Debug']) {$DebugPreference = "Continue"}
    Set-PSDebug -Strict
	Set-StrictMode version Latest
    
    import-module VCommander
    
$index = $remoteComputerName.IndexOf(":")
	if ($index -eq -1) {
		$computeNameNoPort = $remoteComputerName
	} else{
		$computeNameNoPort = $remoteComputerName.Substring(0,$index)
	}
	
    Write-Debug "Checking connectivity to $remoteComputerName"
    $veeamServerConnectionResult = Test-Connection -computer $computeNameNoPort -quiet
    if (!$veeamServerConnectionResult) {
        Write-Error "The Veeam server $remoteComputerName could not be reached."
        Throw The Veeam server $remoteComputerName could not be reached. Connectivity to the specified server could not be established.
    }
      
    Write-Debug "Checking if Veeam credential file exists."
    if (!(Test-Path $veeamCredential)) {
        Write-Error "Veeam credential file does not exist at $veeamCredential"
        Throw "Veeam credential file does not exists at $veeamCredential. Please check if this file exists at this location.";
    }

    #Decrypt the credentials to access Veeam server
    $cred = New-DecryptCredential -keyFilePath $veeamCredential
    
    Write-Debug "VM Name: $vmName | VM Remote Id: $vmRemoteId | jobName: $jobName"
   
    try {
        $invokeResult = invoke-Command -ErrorAction Stop -ComputerName $remoteComputerName -Credential $cred -Argumentlist $vmName, $vmRemoteId, $jobName, $PSBoundParameters['Debug'] -scriptblock {
    		param ($vmName, $vmRemoteId, $jobName, $debug)
            if ($debug) {$DebugPreference = "Continue"}            

    		Add-PSSnapin -name VeeamPSSnapIn
            
            #If a job name is not specified, we will throw error
    		if ([string]::IsNullOrEmpty($jobName)) {
                Write-Error "Job name was not specified." 
                throw "Job name was not specified."
    		}		
        
            Write-Debug "Looking up job: $jobName"
            $job = Get-VBRJob -WarningAction:SilentlyContinue -Name $jobName
            if ($job -eq $null) {
                Write-Error "Job '$jobName' does not exist. Please add job to Veeam."
                throw "Job '$jobName' does not exist. Please add job to Veeam."
            }
    		
            Write-Debug "Querying for VM in job"  
            $objectsInJob = Get-VBRJobObject -WarningAction:SilentlyContinue -Job $job -Name $vmName
            $correctObjectInJob = $objectsInJob
            if( ($objectsInJob | measure).Count -gt 0) {  #more than one VM with the same name
                $correctObjectInJob = $objectsInJob | where {$_.Object.ObjectId -eq $vmRemoteId} #find the correct one
            }
     
            if( ($correctObjectInJob | measure).Count -gt 0) {
                Write-Host "Staring to remove VM '$vmName' ($vmRemoteId) from job '$jobName'"
                Remove-VBRJobObject -WarningAction:SilentlyContinue -Objects $objectsInJob
                Write-Host "VM '$vmName' ($vmRemoteId) was removed from job '$jobName'" 
                return;
            }
            
            Write-Warning "VM '$vmName' ($vmRemoteId) was not removed because it is not in the job '$jobName'."
        }
    } catch {
        Write-Error "$($_.Exception.Message)" 
        throw $_.Exception
    }     
}


<#
.SYNOPSIS
Performs an on-demand incremental backup for a specific VM. The VM must be added to a backup job and have at least one full backup.

.DESCRIPTION
Performs an on-demand incremental backup for a specific VM. The VM must be added to a backup job and have at least one full backup. This feature is only applicable for Veeam Backup and Replication v8.0 and above.

.EXAMPLE
Start-QuickBackup -remoteComputerName 'veeam.mycompany.com' -veeamCredential 'C:\\Scripts\veeam_encrypted_key.xml' -vmName 'VM001' -vmRemoteId 'vm-11538' -hostName 'esx1.embotics.com' -jobName 'Monthly Backup'

This performs a quick backup of the specified VM in the job 'Monthly Backup'.
#>
function Start-QuickBackup {
    [CmdletBinding()]
	param(
        [string] $remoteComputerName = $(Throw "Provide the computer name of the Veeam server."), 
        [String] $veeamCredential = $(Throw "Provide the credential to access Veeam server."),
        [string] $hostName = $(Throw "Provide the host where the VM is located at."), 
        [string] $vmName = $(Throw "Provide the VM name."), 
        [string] $vmRemoteId = $(Throw "Provide the remote ID of this VM."),
		[string] $jobName = $(Throw "Enter a job name.")
    )
    if ($PSBoundParameters['Debug']) {$DebugPreference = "Continue"}
    Set-PSDebug -Strict
	Set-StrictMode version Latest
    
    import-module VCommander

	$index = $remoteComputerName.IndexOf(":")
	if ($index -eq -1) {
		$computeNameNoPort = $remoteComputerName
	} else{
		$computeNameNoPort = $remoteComputerName.Substring(0,$index)
	}

    Write-Debug "Checking connectivity to $remoteComputerName"
    $veeamServerConnectionResult = Test-Connection -computer $computeNameNoPort -quiet
    if (!$veeamServerConnectionResult) {
        Write-Error "The Veeam server $remoteComputerName could not be reached."
        Throw The Veeam server $remoteComputerName could not be reached. Connectivity to the specified server could not be established.
    }
      
    Write-Debug "Checking if Veeam credential file exists."
    if (!(Test-Path $veeamCredential)) {
        Write-Error "Veeam credential file does not exist at $veeamCredential"
        Throw "Veeam credential file does not exists at $veeamCredential. Please check if this file exists at this location.";
    }
    
    #Decrypt the credentials to access Veeam server
    $cred = New-DecryptCredential -keyFilePath $veeamCredential
    
    Write-Debug "host: $hostName | VM Name: $vmName | VM Remote Id: $vmRemoteId | jobName: $jobName"      
    try {
        $invokeResult = invoke-Command -ErrorAction Stop -ComputerName $remoteComputerName -Credential $cred -Argumentlist $hostName, $vmName, $vmRemoteId, $jobName, $PSBoundParameters['Debug'] -scriptblock {
    		param ($hostName, $vmName, $vmRemoteId, $jobName, $debug)
            if ($debug) {$DebugPreference = "Continue"}  

    		Add-PSSnapin -name VeeamPSSnapIn
            
             #If a job name is not specified, we will throw error
    		if ([string]::IsNullOrEmpty($jobName)) {
                Write-Error "Job name was not specified." 
                throw "Job name was not specified."
    		}
       	
    		Write-Debug "Querying for host"
    		$serverObject = Get-VBRServer -WarningAction:SilentlyContinue -Name $hostName
            if ($serverobject -eq $null) {                
                Write-Error "Host '$hostName' does not exist or has not been added to Veeam."
                throw "Host '$hostName' does not exist or has not been added to Veeam. Please verify this host exists or added to Veeam."
            }
           
            Write-Debug "Looking up job: $jobName"
            $job = Get-VBRJob -WarningAction:SilentlyContinue -Name $jobName
            if ($job -eq $null) {
                Write-Error "Job '$jobName' does not exist. Please add job to Veeam."
                throw "Job '$jobName' does not exist. Please add job to Veeam."
            }
              
            #Check if VM already in job
            Write-Debug "Checking if VM is in job"  
            $objectsInJob = Get-VBRJobObject -WarningAction:SilentlyContinue -Job $job -Name $vmName
            $correctObjectInJob = $objectsInJob
            if( ($objectsInJob | measure).Count -gt 0) {  #more than one VM with the same name
                $correctObjectInJob = $objectsInJob | where {$_.Object.ObjectId -eq $vmRemoteId} #find the correct one
            }
            
            if ($correctObjectInJob -eq $null) {
                throw "VM '$vmName' ($vmRemoteId) does not exist in job '$jobName'"
            }
            
            Write-Debug "Querying for VM"	
    		$jobObject = Find-VBRViEntity -Name $vmName -WarningAction:SilentlyContinue | where {$_.Reference -eq $vmRemoteId}

			if ($jobObject -eq $null) {
                 throw "VM '$vmName' with remote id '$vmRemoteId' does not exist. Please verify that this VM exists in one of the vCenters managed by Veeam.";
            }
           
            Write-Debug "Performing quick backup of VM '$vmName' ($vmRemoteId)"
            $result = Start-VBRQuickBackup -WarningAction:SilentlyContinue -VM $jobObject
			if ($result -eq $null) {
				throw "Unable to perform quick backup of VM '$vmName'. See Veeam log for details."
            }

			Write-Debug "$vmName backed up"
        } 
    } catch {
        Write-Error "$($_.Exception.Message)" 
        throw $_.Exception
    }
}